/*-
|> File
|| decode11.c
||
|> Description
||
|> Comments
|| There are 4 types of things in X11: requests, replies, errors, and events.
|| 
|| Each of them has a format defined by a small integer that defines
|| the type of the thing.
|| 
|| Requests have an opcode in the first byte.
|| Events have a code in the first byte.
|| Errors have a code in the second byte (the first byte is 0)
|| Replies ...
|| 
|| Replies have a sequence number in bytes 2 and 3.  The sequence
|| number should be used to identify the request that was sent, and
|| from that request we can determine the type of the reply.
||
|> History
||
||==================================================================|>
*/
#ifndef _NOIDENT
#ident "$Id: decode11.c,v 2.4 1995/10/04 20:56:17 mreynold Exp $"
#endif

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#define NEED_REPLIES
#include <X11/Xproto.h>
#include <X11/Xlib.h>

#include "printer.h"
#include "printlib.h"
#include "printlibP.h"

#include "global.h"

/*- Declare initialized globals. */
/* standard GC created with all default values */
XGCValues               initial_GC =
{
  GXcopy,                       /* function */
  AllPlanes,                    /* plane_mask */
  0L,                           /* foreground */
  1L,                           /* background */
  0,                            /* line_width */
  LineSolid,                    /* line_style */
  CapButt,                      /* cap_style */
  JoinMiter,                    /* join_style */
  FillSolid,                    /* fill_style */
  EvenOddRule,                  /* fill_rule */
  ArcPieSlice,                  /* arc_mode */
  ~0L,                          /* tile, impossible (unknown) resource */
  ~0L,                          /* stipple, impossible (unknown) resource */
  0,                            /* ts_x_origin */
  0,                            /* ts_y_origin */
  ~0L,                          /* font, impossible (unknown) resource */
  ClipByChildren,               /* subwindow_mode */
  True,                         /* graphics_exposures */
  0,                            /* clip_x_origin */
  0,                            /* clip_y_origin */
  None,                         /* clip_mask */
  0,                            /* dash_offset */
  4                             /* dashes (list [4,4]) */
};

XWindowAttributes       initial_Attrs =
{
  0,                            /* x */
  0,                            /* y */
  0,                            /* width */
  0,                            /* height */
  0,                            /* border width */
  8,                            /* default depth */
  0,                            /* visual */
  0,                            /* root window of containing screen */
  InputOutput,                  /* class */
  0,                            /* bit gravity */
  0,                            /* win gravity */
  NotUseful,                    /* backing store */
  0,                            /* backing pixel */
  False,                        /* save under */
  0,                            /* colormap */
  False,                        /* colormap installed flag */
  IsUnmapped,                   /* map state */
  0,                            /* all event mask */
  0,                            /* my event mask */
  0,                            /* do-not-propogate mask */
  False,                        /* override-redirect */
  0                             /* screen pointer */
};

int                     THE_OPCODE = 144;

int                     HandleX_Request (
                                          void *const buf, unsigned char *newmsg, int *newlen,
                                          const int request, const FD fd);

/* function for reading color correction table */
static int
readCorrectionTable (
				 char filename[],
         int *ncorr,
         float ***corrects)

{
  FILE           *fp;
  int             ii;

  if (!(fp = fopen (filename, "r")))
  {
    fprintf (stderr, "Failed to open color correction file %s.\n", filename);
    return 0;
  }
  fscanf (fp, "%d", ncorr);
  *corrects = (float **) malloc (sizeof (float *) * (*ncorr));

  for (ii = 0; ii < *ncorr; ii++)
    (*corrects)[ii] = (float *) malloc (3 * sizeof (float));

  for (ii = 0; ii < *ncorr; ii++)
    if (fscanf (fp, "%f %f %f\n",
          &((*corrects)[ii][0]),
          &((*corrects)[ii][1]),
          &((*corrects)[ii][2])) != 3)
    {
      fprintf (stderr, "Failed to read color correction file %s.\n",
         filename);
      fclose (fp);
      return 0;
    }

  fclose (fp);
  return -1;
}

/*-=================================================================|>
||
|> Public functions
||
||==================================================================|>
||
|> Function
|| DecodeRequest
||
|> Description
|| return TRUE/FALSE as to whether this request should be
|| forwarded to the server.
||
|> Comments
|| None.
||
|> Parameters
|| -void.
||
|> Returns
|| Returns void.
||
|> Exits
|| Exits if:
||
|> Limits
||   
||
||==================================================================|>
*/

Bool
DecodeRequest (
                const FD fd,
                unsigned char *const buf,
                unsigned char *const newmsg,
                int *const newlen)

{
  const short             request = buf ? *buf : 0;
  xGenericReply          *reply = (xGenericReply *) newmsg;

	/* Sanity check. */
  if (!buf || !newmsg || !newlen)
	{
    fprintf (stderr, "Bad arguments to DecodeRequest!\n");
		return 0;
  }

  *newlen = 0;

/* set reply sequence # in advance */
  reply->sequenceNumber = ++(PCS[fd].sequence),
    reply->type = X_Reply;

  if (debug)
    printf ("Request opcode = %d.\n", request);

  if ((request <= 0) || (127 < request))
  {
    xOpenPrintReq          *openReq;
    xClosePrintReq         *closeReq;
    xSetPrintReq           *setReq;
    xPrintReply            *printRply = (xPrintReply *) newmsg;

    if (*buf == THE_OPCODE)
    {
      printRply->type = 1, printRply->length = 0;
      switch (buf[1])
      {
  
/* print showpage and close the file */

      case X_ClosePrint:
        closeReq = (xClosePrintReq *) buf;
        if (PCS[fd].printfp)
        {
          fprintf (PCS[fd].printfp, "\nshowpage\n");
          if (PCS[fd].printDirect)
            pclose (PCS[fd].printfp);
          else
            fclose (PCS[fd].printfp);

          PCS[fd].printfp = 0;
        }
        break;

      case X_OpenPrint:

/* read the print request */
        openReq = (xOpenPrintReq *) buf;
        PCS[fd].printfile = (char *) malloc (openReq->strlength + 1);
        PCS[fd].correctFile = 0;
        strncpy (PCS[fd].printfile, openReq->filename,
                 openReq->strlength);
        PCS[fd].printfile[openReq->strlength] = '\0';

/* if he requested direct printing, then the filename is the name of the
   printer */

        if (PCS[fd].printDirect)
        {
          char                    cmd[0x100] = {0};

          sprintf (cmd, "lpr -P %s", PCS[fd].printfile);
          if (!(PCS[fd].printfp = popen (cmd, "w")))
            fprintf (stderr, "failed to open pipe %s\n", cmd);
        }
        else if ((PCS[fd].printfp =
                  fopen (PCS[fd].printfile, "w")) == NULL)
          fprintf (stderr, "Failed to open file %s.\n",
                   PCS[fd].printfile);

        printRply->flag = (PCS[fd].printfp != NULL);
        PCS[fd].win = openReq->drawable;

/* put in the postscript header information */

        if (printRply->flag) {
          initprocs (&PCS[fd], fd);
	}
        *newlen = sizeof (printRply);
        break;

/* for setting printer attributes, can be extended */

      case X_SetPrint:
        setReq = (xSetPrintReq *) buf;
        PCS[fd].preserveAspect = setReq->aspect;
        PCS[fd].fontScale = setReq->fontScale;
        PCS[fd].pixmapScale = setReq->pixmapScale;
        PCS[fd].pageWidth = setReq->pageWidth;
        PCS[fd].pageHeight = setReq->pageHeight;
        PCS[fd].printDirect = setReq->printDirect;
        PCS[fd].landscape = setReq->landscape;

/* copy the name of the color correction file and load the values, if there
   is one */

        if (setReq->strlength)
        {
          PCS[fd].correctFile = malloc (setReq->strlength + 1);
          strncpy (PCS[fd].correctFile, setReq->correctFile,
                   setReq->strlength);
          PCS[fd].correctFile[setReq->strlength] = '\0';

/* if it failed, then set the correction file to NULL */

          if (!readCorrectionTable (PCS[fd].correctFile,
                                    &PCS[fd].numCorrect,
                                    &PCS[fd].corrects))
          {
            fprintf (stderr, "failed to read correction table %s\n", 
		     PCS[ fd].correctFile);
            free (PCS[fd].correctFile);
            PCS[fd].correctFile = 0;
          }
        }

/* open up the prolog file, which is a cheap excuse to do setpagedevice, and
   sprintf the requested page size to the string, and write it to the output
   file */

	if (setReq->prologlength) 
	{
	  PCS[fd].prologFile = malloc( setReq->prologlength + 1);
	  strncpy( PCS[ fd].prologFile, setReq->prologFile, 
		   setReq->prologlength);
	  PCS[ fd].prologFile[ setReq->prologlength] = '\0';
	} 
	else 
	{
	  PCS[ fd].prologFile = NULL;
	}
        break;
      default:
        fprintf (stderr, "Received erroneous minor opcode.\n");
      }
    }

    return True;                /* maybe this will be useful in the future */
  }

  /*- If it was not handled above, handle it here. */
  return (HandleX_Request (buf, newmsg, newlen, request, fd));
}

/*- end decode11.c
*/
